#
# $Id: ndk_manager.rb,v 1.18 2004/04/22 02:32:54 ko1 Exp $
# Create : K.S. 04/04/17 17:00:44
#

require 'rice/irc'
require 'ndk_err'
require 'ndk_config'
require 'ndk_state'
require 'ndk_client'

module Nadoka
  Cmd = ::RICE::Command
  Rpl = ::RICE::Reply

  require 'ndk_bot'
  
  class NDK_Manager
    def initialize rc
      @rc = rc
      @clients= []
      @prev_timer = Time.now

      @state   = NDK_State.new self
      reload_config
      @state.config = @config

      host, port, @server_passwd = next_server_info
      
      @server = ::RICE::Connection.new(host, port)
      @connected = false
    end
    attr_reader :state, :connected
    
    def client_count
      @clients.size
    end
    
    def next_server_info
      svinfo = @config.server_list.shift
      @config.server_list.push svinfo
      [svinfo[:host], svinfo[:port], svinfo[:pass]]
    end
    
    def reload_config
      @config = NDK_Config.new(self, @rc)
    end

    def start_server_thread
      @server.regist{|rq, wq|
        Thread.stop
        @rq = rq
        begin
          @connected = false
          server_main_proc
        rescue Exception
          ndk_error $!
          @server_thread.kill  if @server_thread  && @server_thread.alive?
          @clients_thread.kill if @clients_thread && @clients_thread.alive?
        end
      }
      @server_thread = Thread.new{
        begin
          @config.slog "Server connection to #{@server.server}:#{@server.port}"
          @server.start(1)
        rescue RICE::Connection::Closed
          host, port, @server_passwd = next_server_info
          @server.server = host
          @server.port   = port
          retry
        rescue Exception
          ndk_error $!
          @clients_thread.kill if @clients_thread && @clients_thread.alive?
        end
      }
    end
    
    def server_main_proc
      # login
      if @server_passwd
        send_to_server Cmd.pass(@server_passwd)
      end
      send_to_server Cmd.nick(@config.nick)
      @state.nick = @config.nick
      
      send_to_server Cmd.user(@config.user,
                              @config.hostname, @config.servername,
                              @config.realname)
      # wait welcome message
      while q = recv_from_server
        if q.command == '001'
          break
        elsif q.command == '433'
          # Nickname is already in use.
          nick = @state.nick_succ
          send_to_server Cmd.nick(nick)
        end
      end
      
      # join to default channels
      @config.default_channels.each{|ch|
        send_to_server Cmd.join(ch)
      }

      @connected = true
      # loop
      while q = recv_from_server
        
        case q.command
        when 'PING'
          send_to_server Cmd.pong(q.params[0])
          next
        when 'PRIVMSG'
          if ctcp_message?(q.params[1])
            ctcp_message(q)
          end
        when 'JOIN'
          @state.on_join(sender_of(q), q.params[0])
        when 'PART'
          @state.on_part(sender_of(q), q.params[0])
        when 'NICK'
          @state.on_nick(sender_of(q), q.params[0])
        when 'QUIT'
          @state.on_quit(sender_of(q), q.params[0])
        when '353'
          @state.on_353(q.params[2], q.params[3])
        else
          
        end
        send_to_clients q
        send_to_bot q
        logging q
      end
    end
    
    def start_clients_thread
      @clients_thread = Thread.new{

        begin
          @cserver = TCPServer.new(@config.client_server_port)
          @config.slog "Open Client Server Port: #{@cserver.addr.join(' ')}"
          
          while true
            # wait for client connections
            Thread.start(@cserver.accept){|cc|
              client = NDK_Client.new(@config, cc, self)
              @clients << client
              client.start
              @clients.delete client
            }
          end
        rescue
          ndk_error $!
        ensure
          @clients.each{|cl|
            cl.kill
          }
          @server_thread.kill if @server_thread.alive?
        end
      }
    end

    def about_me?(msg)
      /^#{@state.nick}!/ =~ msg.prefix
    end
    
    def start
      start_server_thread
      start_clients_thread

      begin
        @server_thread.join
      rescue NDK_QuitProgram
        exit
      rescue Interrupt
        @server_thread.kill  if @server_thread.alive?
        @clients_thread.kill if @clients_thread.alive?
      ensure
        begin
          @cserver.close if @cserver
        rescue IOError
        end
      end
    end
    
    def send_to_server msg
      @config.dlog "[>S] #{msg}"
      @server << msg
    end
    
    def recv_from_server
      while q = @rq.pop
        # Timer event?
        t = Time.now
        if (t.to_i - @prev_timer.to_i) > 60
          @prev_timer = t
          send_to_bot nil, t
        end

        ##
        case q.command
        when 'PING'
          @server << Cmd.pong(q.params[0])
        when 'PONG'
          # ignore
        else
          @config.dlog "[<S] #{q}"
          return q
        end
      end
    end

    # server -> clients
    def send_to_clients msg
      @clients.each{|cl|
        cl << msg
      }
    end

    # clientA -> other clients
    # bot     -> clients
    def send_to_clients_otherwise msg, elt
      @clients.each{|cl|
        if cl != elt
          cl.add_prefix(msg) unless msg.prefix
          cl << msg
        end
      }
      send_to_bot msg if elt
      logging msg
    end
    
    def join_channel ch
      send_to_server Cmd.topic(ch)
      send_to_server Cmd.names(ch)
    end

    def ctcp_message? arg
      arg[0] == 1
    end
    
    def ctcp_message msg
      if /\001(.+)\001/ =~ msg.params[1]
        ctcp_cmd = $1
        case ctcp_cmd
        when 'VERSION'
          send_to_server Cmd.notice(sender_of(msg), "\001VERSION #{Nadoka.version}\001")
        when 'TIME'
          send_to_server Cmd.notice(sender_of(msg), "\001TIME #{Time.now}\001")
        else
          
        end
      end
    end
    
    def sender_of msg
      if /^([^!]+)\!?/ =~ msg.prefix.to_s
        $1
      else
        @state.nick
      end
    end

    # dispatch to bots
    def send_to_bot msg, arg = nil

      selector = 'on_' +
      if msg.respond_to? :command
        if /^\d+$/ =~ msg.command
          RICE::Reply::Replies_num_to_name[msg.command]
        else 
          msg.command.downcase
        end
      else
        msg.to_s
      end
      
      @config.bots.each{|bot|
        begin
          if bot.respond_to? selector
            unless msg.respond_to? :command
              bot.raw_prefix = nil
              bot.__send__(selector, arg)
            else
              bot.raw_prefix = msg.prefix
              bot.__send__(selector, sender_of(msg), *msg.params)
            end
          end
          
        rescue NDK_BotBreak
          break
          
        rescue Exception
          ndk_error $!
        end
      }
    end

    # logging
    def logging msg
      user = sender_of(msg)
      ch   = msg.params[0]
      case msg.command
      when 'PRIVMSG'
        str = "<#{ch}:#{user}> #{msg.params[1]}"
        @config.clog ch, str
        
      when 'NOTICE'
        str = "{#{ch}:#{user}} #{msg.params[1]}"
        @config.clog ch, str

      when 'TOPIC'
        str = "{#{ch}:#{user}} #{msg.params[1]}"
        @config.clog ch, str
        
      when 'JOIN', 'PART', 'NICK', 'QUIT'
        # ignore
        
      when /^\d+/
        str = msg.command + ' ' + msg.params.join(' ')
        @config.log str
        
      else
        # other command
        @config.log msg.to_s
      end
    end

    def ndk_error err
      $stderr.puts err
      @config.slog err
      @config.slog err.backtrace.join(' ')
    end
    
  end
end


